// Traitement du CAN utilisant la couche HAL du noyau STM32

#define CAN_RX_QUEUE_SIZE 16

// Les typedef sont dans le fichier principal


// Variables driver
CAN_HandleTypeDef hCAN;
CAN_RxHeaderTypeDef RxMsgHeader;
uint32_t pTxMailbox ;
CanMsg can_rx_queue[CAN_RX_QUEUE_SIZE];
volatile uint8_t can_rx_head;
volatile uint8_t can_rx_tail;
volatile uint8_t can_rx_lost;

void can_rx_read(CAN_HandleTypeDef *hcan);

// Gestion du FIFO de lecture (plus grande capacité que le FIFO du hardware)
void can_rx_queue_clear(void)
{
  __HAL_CAN_DISABLE_IT(&hCAN, CAN_IT_RX_FIFO0_MSG_PENDING);
  __asm volatile( "dsb" );            // Synchronization of the pipeline to guarantee
  __asm volatile( "isb" );            // that the protection against interrupts is effective
  can_rx_head = can_rx_tail = can_rx_lost = 0;          // JMD 2017/07/18
  __HAL_CAN_ENABLE_IT(&hCAN, CAN_IT_RX_FIFO0_MSG_PENDING);
}

uint8_t can_rx_available(void)
{
  if ( can_rx_head >= can_rx_tail )
    return can_rx_head - can_rx_tail ;
  else
    return CAN_RX_QUEUE_SIZE - ( can_rx_tail - can_rx_head ) ;
}

uint8_t can_frame_lost(void)
{
  if ( can_rx_lost != 0 )
  {
    can_rx_lost = 0 ;
    return 1 ;
  }
  return 0 ;
}

void can_rx_read(CAN_HandleTypeDef *hcan)
{
  CAN_RxHeaderTypeDef RxMsgHeader;
  uint8_t Data[8];      // Poubelle pour la trame à jeter si débordement du grand FIFO

  if (can_rx_available() < CAN_RX_QUEUE_SIZE)   // read the message
  {
    CanMsg* msg = &can_rx_queue[can_rx_head];
//  digitalWrite(LED_JAUNE,HIGH);
    HAL_CAN_GetRxMessage(hcan, 0, &RxMsgHeader, msg->Data);
    // Recopie de l'en-tête dans la structure de type msg
    msg->IDE = RxMsgHeader.IDE ;
    
    if ( msg->IDE == CAN_ID_STD )
      msg->ID = RxMsgHeader.StdId ;
    else
      msg->ID = RxMsgHeader.ExtId ;
      
    msg->RTR = RxMsgHeader.RTR ;
    msg->DLC = RxMsgHeader.DLC ;

    can_rx_head = (can_rx_head == (CAN_RX_QUEUE_SIZE - 1)) ? 0 : (can_rx_head + 1);
//  digitalWrite(LED_JAUNE,LOW);
  }
  else
  {
    HAL_CAN_GetRxMessage(&hCAN, 0, &RxMsgHeader, Data);
    can_rx_lost = 1;            // no place in queue, ignore package
  }
}

CanMsg* can_rx_queue_get(void)
{
  if (can_rx_head == can_rx_tail)     // JMD 2017/07/18 -- changed
    return NULL;
  return &(can_rx_queue[can_rx_tail]);
}


void can_rx_queue_free(void)
{
  if (can_rx_head != can_rx_tail)     // JMD 2017/07/18 -- changed
  {
    __HAL_CAN_DISABLE_IT(&hCAN, CAN_IT_RX_FIFO0_MSG_PENDING);
    __asm volatile( "dsb" );            // Synchronization of the pipeline to guarantee
    __asm volatile( "isb" );            // that the protection against interrupts is effective
    can_rx_tail = (can_rx_tail == (CAN_RX_QUEUE_SIZE - 1)) ? 0 : (can_rx_tail + 1);
    __HAL_CAN_ENABLE_IT(&hCAN, CAN_IT_RX_FIFO0_MSG_PENDING);
  }
}

 
// Initialisation du CAN
void CANSetup(void)
{
  CAN_FilterTypeDef sFilterConfig = {0} ; //declare CAN filter structure, vide

  hCAN.State = HAL_CAN_STATE_RESET ;
  if ( Serial )
    Serial.end();
  can_rx_head = can_rx_tail = can_rx_lost = 0;

  hCAN.Instance = CAN1;
  hCAN.Init.Prescaler = 16;    // 125 000 b/s
  hCAN.Init.Mode = CAN_MODE_NORMAL;
  hCAN.Init.SyncJumpWidth = CAN_SJW_4TQ;
  hCAN.Init.TimeSeg1 = CAN_BS1_12TQ;
  hCAN.Init.TimeSeg2 = CAN_BS2_5TQ;
  hCAN.Init.TimeTriggeredMode = DISABLE;
  hCAN.Init.AutoBusOff = DISABLE;
  hCAN.Init.AutoWakeUp = DISABLE;
  hCAN.Init.AutoRetransmission = ENABLE;
  hCAN.Init.ReceiveFifoLocked = DISABLE;
  hCAN.Init.TransmitFifoPriority = DISABLE;
    
  if (HAL_CAN_Init(&hCAN) != HAL_OK)
  {
    digitalWrite(LED_JAUNE,HIGH);
    delay ( 200 ) ;    
    digitalWrite(LED_JAUNE,LOW);
  }

  sFilterConfig.FilterFIFOAssignment=CAN_FILTER_FIFO0; //set fifo assignment
  sFilterConfig.FilterIdHigh = 0 ;
  sFilterConfig.FilterIdLow = 0 ;
  sFilterConfig.FilterMaskIdHigh = 0;
  sFilterConfig.FilterMaskIdLow = 0;
  sFilterConfig.FilterScale = CAN_FILTERSCALE_32BIT; //set filter scale
  sFilterConfig.FilterActivation = ENABLE;
  sFilterConfig.FilterBank = 0 ;
  sFilterConfig.FilterMode = CAN_FILTERMODE_IDMASK ;
    
  HAL_CAN_ConfigFilter(&hCAN, &sFilterConfig); //configure CAN filter

  HAL_CAN_ActivateNotification(&hCAN, CAN_IT_RX_FIFO0_MSG_PENDING); //enable interrupts

  if (HAL_CAN_Start(&hCAN) != HAL_OK)
  {
    digitalWrite(LED_JAUNE,HIGH);
    delay ( 200 ) ;    
    digitalWrite(LED_JAUNE,LOW);
    delay ( 300 ) ;    
    digitalWrite(LED_JAUNE,HIGH);
    delay ( 200 ) ;    
    digitalWrite(LED_JAUNE,LOW);
  }
}

// Initialisation bas niveau
void HAL_CAN_MspInit(CAN_HandleTypeDef *hcan)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0} ;

  __HAL_RCC_CAN1_CLK_ENABLE() ;
  __HAL_RCC_GPIOB_CLK_ENABLE();
  
  // Configuration des pins PB8 (RX) et PB9 (TX)
  GPIO_InitStruct.Pin = GPIO_PIN_9;
  GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_HIGH;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  GPIO_InitStruct.Pin = GPIO_PIN_8;
  GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  __HAL_AFIO_REMAP_CAN1_2() ;

  HAL_NVIC_SetPriority(USB_LP_CAN1_RX0_IRQn, 0 , 0) ;
  HAL_NVIC_EnableIRQ(USB_LP_CAN1_RX0_IRQn) ;
}

// Callback de message entrant
void HAL_CAN_RxFifo0MsgPendingCallback (CAN_HandleTypeDef *hcan0)
{
  // envoie la trame reçue dans la file d'attente
  can_rx_read(hcan0) ;
}

// La routine d'interruption appelle la fonction prédéfinie qui détermine la cause de
// l'interruption et gère l'effacement de la demande d'interruption
// Cette routine appellera le callback ci-dessus si interruption de réception
//extern "C" void USB_LP_CAN1_RX0_IRQHandler (void) ;
void USB_LP_CAN1_RX0_IRQHandler (void)
{
  HAL_CAN_IRQHandler(&hCAN) ;
}

// Emission d'une trame. Si aucune mailbox libre, .send retourne avec état CAN_TX_NO_MBX
// Répète .send tant que cet état continue
void CANsend(CanMsg *pmsg)
{
CAN_TxHeaderTypeDef TxMsgHeader;

  // Recopie du paramètre d'entrée dans la structure de type HAL
  TxMsgHeader.StdId = pmsg->ID ;
  TxMsgHeader.ExtId = pmsg->ID ;
  TxMsgHeader.IDE = pmsg->IDE ;
  TxMsgHeader.RTR = pmsg->RTR ;
  TxMsgHeader.DLC = pmsg->DLC ;
  HAL_CAN_AddTxMessage(&hCAN, &TxMsgHeader, pmsg->Data, &pTxMailbox) ;
  delay( 1 ) ; 
}
